@@ -1,5 +1,6 @@
 /* OpenMP directive matching and resolving.
-   Copyright (C) 2005, 2006, 2007 Free Software Foundation, Inc.
+   Copyright (C) 2005, 2006, 2007
+   Free Software Foundation, Inc.
    Contributed by Jakub Jelinek
 
 This file is part of GCC.
@@ -18,7 +19,6 @@ You should have received a copy of the GNU General Public License
 along with GCC; see the file COPYING3.  If not see
 <http://www.gnu.org/licenses/>.  */
 
-
 #include "config.h"
 #include "system.h"
 #include "flags.h"
@@ -409,6 +409,7 @@ gfc_match_omp_parallel (void)
   return MATCH_YES;
 }
 
+
 match
 gfc_match_omp_critical (void)
 {
@@ -423,6 +424,7 @@ gfc_match_omp_critical (void)
   return MATCH_YES;
 }
 
+
 match
 gfc_match_omp_do (void)
 {
@@ -434,6 +436,7 @@ gfc_match_omp_do (void)
   return MATCH_YES;
 }
 
+
 match
 gfc_match_omp_flush (void)
 {
@@ -449,6 +452,7 @@ gfc_match_omp_flush (void)
   return MATCH_YES;
 }
 
+
 match
 gfc_match_omp_threadprivate (void)
 {
@@ -464,12 +468,6 @@ gfc_match_omp_threadprivate (void)
   if (m != MATCH_YES)
     return m;
 
-  if (!targetm.have_tls)
-    {
-      sorry ("threadprivate variables not supported in this target");
-      goto cleanup;
-    }
-
   for (;;)
     {
       m = gfc_match_symbol (&sym, 0);
@@ -477,8 +475,8 @@ gfc_match_omp_threadprivate (void)
 	{
 	case MATCH_YES:
 	  if (sym->attr.in_common)
-	    gfc_error_now ("Threadprivate variable at %C is an element of"
-			   " a COMMON block");
+	    gfc_error_now ("Threadprivate variable at %C is an element of "
+			   "a COMMON block");
 	  else if (gfc_add_threadprivate (&sym->attr, sym->name,
 		   &sym->declared_at) == FAILURE)
 	    goto cleanup;
@@ -524,6 +522,7 @@ cleanup:
   return MATCH_ERROR;
 }
 
+
 match
 gfc_match_omp_parallel_do (void)
 {
@@ -536,6 +535,7 @@ gfc_match_omp_parallel_do (void)
   return MATCH_YES;
 }
 
+
 match
 gfc_match_omp_parallel_sections (void)
 {
@@ -548,6 +548,7 @@ gfc_match_omp_parallel_sections (void)
   return MATCH_YES;
 }
 
+
 match
 gfc_match_omp_parallel_workshare (void)
 {
@@ -559,6 +560,7 @@ gfc_match_omp_parallel_workshare (void)
   return MATCH_YES;
 }
 
+
 match
 gfc_match_omp_sections (void)
 {
@@ -570,6 +572,7 @@ gfc_match_omp_sections (void)
   return MATCH_YES;
 }
 
+
 match
 gfc_match_omp_single (void)
 {
@@ -582,6 +585,7 @@ gfc_match_omp_single (void)
   return MATCH_YES;
 }
 
+
 match
 gfc_match_omp_workshare (void)
 {
@@ -592,6 +596,7 @@ gfc_match_omp_workshare (void)
   return MATCH_YES;
 }
 
+
 match
 gfc_match_omp_master (void)
 {
@@ -602,6 +607,7 @@ gfc_match_omp_master (void)
   return MATCH_YES;
 }
 
+
 match
 gfc_match_omp_ordered (void)
 {
@@ -612,6 +618,7 @@ gfc_match_omp_ordered (void)
   return MATCH_YES;
 }
 
+
 match
 gfc_match_omp_atomic (void)
 {
@@ -622,6 +629,7 @@ gfc_match_omp_atomic (void)
   return MATCH_YES;
 }
 
+
 match
 gfc_match_omp_barrier (void)
 {
@@ -632,6 +640,7 @@ gfc_match_omp_barrier (void)
   return MATCH_YES;
 }
 
+
 match
 gfc_match_omp_end_nowait (void)
 {
@@ -645,6 +654,7 @@ gfc_match_omp_end_nowait (void)
   return MATCH_YES;
 }
 
+
 match
 gfc_match_omp_end_single (void)
 {
@@ -662,6 +672,7 @@ gfc_match_omp_end_single (void)
   return MATCH_YES;
 }
 
+
 /* OpenMP directive resolving routines.  */
 
 static void
@@ -690,23 +701,57 @@ resolve_omp_clauses (gfc_code *code)
       gfc_expr *expr = omp_clauses->num_threads;
       if (gfc_resolve_expr (expr) == FAILURE
 	  || expr->ts.type != BT_INTEGER || expr->rank != 0)
-	gfc_error ("NUM_THREADS clause at %L requires a scalar"
-		   " INTEGER expression", &expr->where);
+	gfc_error ("NUM_THREADS clause at %L requires a scalar "
+		   "INTEGER expression", &expr->where);
     }
   if (omp_clauses->chunk_size)
     {
       gfc_expr *expr = omp_clauses->chunk_size;
       if (gfc_resolve_expr (expr) == FAILURE
 	  || expr->ts.type != BT_INTEGER || expr->rank != 0)
-	gfc_error ("SCHEDULE clause's chunk_size at %L requires"
-		   " a scalar INTEGER expression", &expr->where);
+	gfc_error ("SCHEDULE clause's chunk_size at %L requires "
+		   "a scalar INTEGER expression", &expr->where);
     }
 
   /* Check that no symbol appears on multiple clauses, except that
      a symbol can appear on both firstprivate and lastprivate.  */
   for (list = 0; list < OMP_LIST_NUM; list++)
     for (n = omp_clauses->lists[list]; n; n = n->next)
-      n->sym->mark = 0;
+      {
+	n->sym->mark = 0;
+	if (n->sym->attr.flavor == FL_VARIABLE)
+	  continue;
+	if (n->sym->attr.flavor == FL_PROCEDURE
+	    && n->sym->result == n->sym
+	    && n->sym->attr.function)
+	  {
+	    if (gfc_current_ns->proc_name == n->sym
+		|| (gfc_current_ns->parent
+		    && gfc_current_ns->parent->proc_name == n->sym))
+	      continue;
+	    if (gfc_current_ns->proc_name->attr.entry_master)
+	      {
+		gfc_entry_list *el = gfc_current_ns->entries;
+		for (; el; el = el->next)
+		  if (el->sym == n->sym)
+		    break;
+		if (el)
+		  continue;
+	      }
+	    if (gfc_current_ns->parent
+		&& gfc_current_ns->parent->proc_name->attr.entry_master)
+	      {
+		gfc_entry_list *el = gfc_current_ns->parent->entries;
+		for (; el; el = el->next)
+		  if (el->sym == n->sym)
+		    break;
+		if (el)
+		  continue;
+	      }
+	  }
+	gfc_error ("Object '%s' is not a variable at %L", n->sym->name,
+		   &code->loc);
+      }
 
   for (list = 0; list < OMP_LIST_NUM; list++)
     if (list != OMP_LIST_FIRSTPRIVATE && list != OMP_LIST_LASTPRIVATE)
@@ -767,25 +812,31 @@ resolve_omp_clauses (gfc_code *code)
 		if (n->sym->attr.allocatable)
 		  gfc_error ("COPYIN clause object '%s' is ALLOCATABLE at %L",
 			     n->sym->name, &code->loc);
+		if (n->sym->ts.type == BT_DERIVED && n->sym->ts.derived->attr.alloc_comp)
+		  gfc_error ("COPYIN clause object '%s' at %L has ALLOCATABLE components",
+			     n->sym->name, &code->loc);
 	      }
 	    break;
 	  case OMP_LIST_COPYPRIVATE:
 	    for (; n != NULL; n = n->next)
 	      {
 		if (n->sym->as && n->sym->as->type == AS_ASSUMED_SIZE)
-		  gfc_error ("Assumed size array '%s' in COPYPRIVATE clause"
-			     " at %L", n->sym->name, &code->loc);
+		  gfc_error ("Assumed size array '%s' in COPYPRIVATE clause "
+			     "at %L", n->sym->name, &code->loc);
 		if (n->sym->attr.allocatable)
-		  gfc_error ("COPYPRIVATE clause object '%s' is ALLOCATABLE"
-			     " at %L", n->sym->name, &code->loc);
+		  gfc_error ("COPYPRIVATE clause object '%s' is ALLOCATABLE "
+			     "at %L", n->sym->name, &code->loc);
+		if (n->sym->ts.type == BT_DERIVED && n->sym->ts.derived->attr.alloc_comp)
+		  gfc_error ("COPYPRIVATE clause object '%s' at %L has ALLOCATABLE components",
+			     n->sym->name, &code->loc);
 	      }
 	    break;
 	  case OMP_LIST_SHARED:
 	    for (; n != NULL; n = n->next)
 	      {
 		if (n->sym->attr.threadprivate)
-		  gfc_error ("THREADPRIVATE object '%s' in SHARED clause at"
-			     " %L", n->sym->name, &code->loc);
+		  gfc_error ("THREADPRIVATE object '%s' in SHARED clause at "
+			     "%L", n->sym->name, &code->loc);
 		if (n->sym->attr.cray_pointee)
 		  gfc_error ("Cray pointee '%s' in SHARED clause at %L",
 			    n->sym->name, &code->loc);
@@ -808,6 +859,11 @@ resolve_omp_clauses (gfc_code *code)
 		    if (n->sym->attr.allocatable)
 		      gfc_error ("%s clause object '%s' is ALLOCATABLE at %L",
 				 name, n->sym->name, &code->loc);
+		    /* Variables in REDUCTION-clauses must be of intrinsic type (flagged below).  */
+		    if ((list < OMP_LIST_REDUCTION_FIRST || list > OMP_LIST_REDUCTION_LAST) &&
+		        n->sym->ts.type == BT_DERIVED && n->sym->ts.derived->attr.alloc_comp)
+		      gfc_error ("%s clause object '%s' has ALLOCATABLE components at %L",
+				 name, n->sym->name, &code->loc);
 		    if (n->sym->attr.cray_pointer)
 		      gfc_error ("Cray pointer '%s' in %s clause at %L",
 				 n->sym->name, name, &code->loc);
@@ -818,8 +874,8 @@ resolve_omp_clauses (gfc_code *code)
 		if (n->sym->attr.in_namelist
 		    && (list < OMP_LIST_REDUCTION_FIRST
 			|| list > OMP_LIST_REDUCTION_LAST))
-		  gfc_error ("Variable '%s' in %s clause is used in"
-			     " NAMELIST statement at %L",
+		  gfc_error ("Variable '%s' in %s clause is used in "
+			     "NAMELIST statement at %L",
 			     n->sym->name, name, &code->loc);
 		switch (list)
 		  {
@@ -827,19 +883,19 @@ resolve_omp_clauses (gfc_code *code)
 		  case OMP_LIST_MULT:
 		  case OMP_LIST_SUB:
 		    if (!gfc_numeric_ts (&n->sym->ts))
-		      gfc_error ("%c REDUCTION variable '%s' is %s at %L",
+		      gfc_error ("%c REDUCTION variable '%s' at %L must be of numeric type, got %s",
 				 list == OMP_LIST_PLUS ? '+'
 				 : list == OMP_LIST_MULT ? '*' : '-',
-				 n->sym->name, gfc_typename (&n->sym->ts),
-				 &code->loc);
+				 n->sym->name, &code->loc,
+				 gfc_typename (&n->sym->ts));
 		    break;
 		  case OMP_LIST_AND:
 		  case OMP_LIST_OR:
 		  case OMP_LIST_EQV:
 		  case OMP_LIST_NEQV:
 		    if (n->sym->ts.type != BT_LOGICAL)
-		      gfc_error ("%s REDUCTION variable '%s' must be LOGICAL"
-				 " at %L",
+		      gfc_error ("%s REDUCTION variable '%s' must be LOGICAL "
+				 "at %L",
 				 list == OMP_LIST_AND ? ".AND."
 				 : list == OMP_LIST_OR ? ".OR."
 				 : list == OMP_LIST_EQV ? ".EQV." : ".NEQV.",
@@ -849,8 +905,8 @@ resolve_omp_clauses (gfc_code *code)
 		  case OMP_LIST_MIN:
 		    if (n->sym->ts.type != BT_INTEGER
 			&& n->sym->ts.type != BT_REAL)
-		      gfc_error ("%s REDUCTION variable '%s' must be"
-				 " INTEGER or REAL at %L",
+		      gfc_error ("%s REDUCTION variable '%s' must be "
+				 "INTEGER or REAL at %L",
 				 list == OMP_LIST_MAX ? "MAX" : "MIN",
 				 n->sym->name, &code->loc);
 		    break;
@@ -858,8 +914,8 @@ resolve_omp_clauses (gfc_code *code)
 		  case OMP_LIST_IOR:
 		  case OMP_LIST_IEOR:
 		    if (n->sym->ts.type != BT_INTEGER)
-		      gfc_error ("%s REDUCTION variable '%s' must be INTEGER"
-				 " at %L",
+		      gfc_error ("%s REDUCTION variable '%s' must be INTEGER "
+				 "at %L",
 				 list == OMP_LIST_IAND ? "IAND"
 				 : list == OMP_LIST_MULT ? "IOR" : "IEOR",
 				 n->sym->name, &code->loc);
@@ -877,6 +933,7 @@ resolve_omp_clauses (gfc_code *code)
       }
 }
 
+
 /* Return true if SYM is ever referenced in EXPR except in the SE node.  */
 
 static bool
@@ -916,6 +973,7 @@ expr_references_sym (gfc_expr *e, gfc_symbol *s, gfc_expr *se)
     }
 }
 
+
 /* If EXPR is a conversion function that widens the type
    if WIDENING is true or narrows the type if WIDENING is false,
    return the inner expression, otherwise return NULL.  */
@@ -928,7 +986,7 @@ is_conversion (gfc_expr *expr, bool widening)
   if (expr->expr_type != EXPR_FUNCTION
       || expr->value.function.isym == NULL
       || expr->value.function.esym != NULL
-      || expr->value.function.isym->generic_id != GFC_ISYM_CONVERSION)
+      || expr->value.function.isym->id != GFC_ISYM_CONVERSION)
     return NULL;
 
   if (widening)
@@ -949,6 +1007,7 @@ is_conversion (gfc_expr *expr, bool widening)
   return NULL;
 }
 
+
 static void
 resolve_omp_atomic (gfc_code *code)
 {
@@ -967,8 +1026,8 @@ resolve_omp_atomic (gfc_code *code)
 	  && code->expr->ts.type != BT_COMPLEX
 	  && code->expr->ts.type != BT_LOGICAL))
     {
-      gfc_error ("!$OMP ATOMIC statement must set a scalar variable of"
-		 " intrinsic type at %L", &code->loc);
+      gfc_error ("!$OMP ATOMIC statement must set a scalar variable of "
+		 "intrinsic type at %L", &code->loc);
       return;
     }
 
@@ -1007,8 +1066,8 @@ resolve_omp_atomic (gfc_code *code)
 	  alt_op = INTRINSIC_EQV;
 	  break;
 	default:
-	  gfc_error ("!$OMP ATOMIC assignment operator must be"
-		     " +, *, -, /, .AND., .OR., .EQV. or .NEQV. at %L",
+	  gfc_error ("!$OMP ATOMIC assignment operator must be "
+		     "+, *, -, /, .AND., .OR., .EQV. or .NEQV. at %L",
 		     &expr2->where);
 	  return;
 	}
@@ -1055,8 +1114,8 @@ resolve_omp_atomic (gfc_code *code)
 
 	  if (v == NULL)
 	    {
-	      gfc_error ("!$OMP ATOMIC assignment must be var = var op expr"
-			 " or var = expr op var at %L", &expr2->where);
+	      gfc_error ("!$OMP ATOMIC assignment must be var = var op expr "
+			 "or var = expr op var at %L", &expr2->where);
 	      return;
 	    }
 
@@ -1069,9 +1128,9 @@ resolve_omp_atomic (gfc_code *code)
 		case INTRINSIC_DIVIDE:
 		case INTRINSIC_EQV:
 		case INTRINSIC_NEQV:
-		  gfc_error ("!$OMP ATOMIC var = var op expr not"
-			     " mathematically equivalent to var = var op"
-			     " (expr) at %L", &expr2->where);
+		  gfc_error ("!$OMP ATOMIC var = var op expr not "
+			     "mathematically equivalent to var = var op "
+			     "(expr) at %L", &expr2->where);
 		  break;
 		default:
 		  break;
@@ -1101,8 +1160,8 @@ resolve_omp_atomic (gfc_code *code)
 
       if (e->rank != 0 || expr_references_sym (code->expr2, var, v))
 	{
-	  gfc_error ("expr in !$OMP ATOMIC assignment var = var op expr"
-		     " must be scalar and cannot reference var at %L",
+	  gfc_error ("expr in !$OMP ATOMIC assignment var = var op expr "
+		     "must be scalar and cannot reference var at %L",
 		     &expr2->where);
 	  return;
 	}
@@ -1115,7 +1174,7 @@ resolve_omp_atomic (gfc_code *code)
     {
       gfc_actual_arglist *arg, *var_arg;
 
-      switch (expr2->value.function.isym->generic_id)
+      switch (expr2->value.function.isym->id)
 	{
 	case GFC_ISYM_MIN:
 	case GFC_ISYM_MAX:
@@ -1125,15 +1184,15 @@ resolve_omp_atomic (gfc_code *code)
 	case GFC_ISYM_IEOR:
 	  if (expr2->value.function.actual->next->next != NULL)
 	    {
-	      gfc_error ("!$OMP ATOMIC assignment intrinsic IAND, IOR"
+	      gfc_error ("!$OMP ATOMIC assignment intrinsic IAND, IOR "
 			 "or IEOR must have two arguments at %L",
 			 &expr2->where);
 	      return;
 	    }
 	  break;
 	default:
-	  gfc_error ("!$OMP ATOMIC assignment intrinsic must be"
-		     " MIN, MAX, IAND, IOR or IEOR at %L",
+	  gfc_error ("!$OMP ATOMIC assignment intrinsic must be "
+		     "MIN, MAX, IAND, IOR or IEOR at %L",
 		     &expr2->where);
 	  return;
 	}
@@ -1148,17 +1207,17 @@ resolve_omp_atomic (gfc_code *code)
 	      && arg->expr->symtree->n.sym == var)
 	    var_arg = arg;
 	  else if (expr_references_sym (arg->expr, var, NULL))
-	    gfc_error ("!$OMP ATOMIC intrinsic arguments except one must not"
-		       " reference '%s' at %L", var->name, &arg->expr->where);
+	    gfc_error ("!$OMP ATOMIC intrinsic arguments except one must not "
+		       "reference '%s' at %L", var->name, &arg->expr->where);
 	  if (arg->expr->rank != 0)
-	    gfc_error ("!$OMP ATOMIC intrinsic arguments must be scalar"
-		       " at %L", &arg->expr->where);
+	    gfc_error ("!$OMP ATOMIC intrinsic arguments must be scalar "
+		       "at %L", &arg->expr->where);
 	}
 
       if (var_arg == NULL)
 	{
-	  gfc_error ("First or last !$OMP ATOMIC intrinsic argument must"
-		     " be '%s' at %L", var->name, &expr2->where);
+	  gfc_error ("First or last !$OMP ATOMIC intrinsic argument must "
+		     "be '%s' at %L", var->name, &expr2->where);
 	  return;
 	}
 
@@ -1175,10 +1234,11 @@ resolve_omp_atomic (gfc_code *code)
 	}
     }
   else
-    gfc_error ("!$OMP ATOMIC assignment must have an operator or intrinsic"
-	       " on right hand side at %L", &expr2->where);
+    gfc_error ("!$OMP ATOMIC assignment must have an operator or intrinsic "
+	       "on right hand side at %L", &expr2->where);
 }
 
+
 struct omp_context
 {
   gfc_code *code;
@@ -1188,6 +1248,7 @@ struct omp_context
 } *omp_current_ctx;
 gfc_code *omp_current_do_code;
 
+
 void
 gfc_resolve_omp_do_blocks (gfc_code *code, gfc_namespace *ns)
 {
@@ -1196,6 +1257,7 @@ gfc_resolve_omp_do_blocks (gfc_code *code, gfc_namespace *ns)
   gfc_resolve_blocks (code->block, ns);
 }
 
+
 void
 gfc_resolve_omp_parallel_blocks (gfc_code *code, gfc_namespace *ns)
 {
@@ -1224,6 +1286,7 @@ gfc_resolve_omp_parallel_blocks (gfc_code *code, gfc_namespace *ns)
   pointer_set_destroy (ctx.private_iterators);
 }
 
+
 /* Note a DO iterator variable.  This is special in !$omp parallel
    construct, where they are predetermined private.  */
 
@@ -1259,6 +1322,7 @@ gfc_resolve_do_iterator (gfc_code *code, gfc_symbol *sym)
     }
 }
 
+
 static void
 resolve_omp_do (gfc_code *code)
 {
@@ -1272,8 +1336,8 @@ resolve_omp_do (gfc_code *code)
 
   do_code = code->block->next;
   if (do_code->op == EXEC_DO_WHILE)
-    gfc_error ("!$OMP DO cannot be a DO WHILE or DO without loop control at %L",
-	       &do_code->loc);
+    gfc_error ("!$OMP DO cannot be a DO WHILE or DO without loop control "
+	       "at %L", &do_code->loc);
   else
     {
       gcc_assert (do_code->op == EXEC_DO);
@@ -1282,22 +1346,23 @@ resolve_omp_do (gfc_code *code)
 		   &do_code->loc);
       dovar = do_code->ext.iterator->var->symtree->n.sym;
       if (dovar->attr.threadprivate)
-	gfc_error ("!$OMP DO iteration variable must not be THREADPRIVATE at %L",
-		   &do_code->loc);
+	gfc_error ("!$OMP DO iteration variable must not be THREADPRIVATE "
+		   "at %L", &do_code->loc);
       if (code->ext.omp_clauses)
 	for (list = 0; list < OMP_LIST_NUM; list++)
 	  if (list != OMP_LIST_PRIVATE && list != OMP_LIST_LASTPRIVATE)
 	    for (n = code->ext.omp_clauses->lists[list]; n; n = n->next)
 	      if (dovar == n->sym)
 		{
-		  gfc_error ("!$OMP DO iteration variable present on clause"
-			     " other than PRIVATE or LASTPRIVATE at %L",
+		  gfc_error ("!$OMP DO iteration variable present on clause "
+			     "other than PRIVATE or LASTPRIVATE at %L",
 			     &do_code->loc);
 		  break;
 		}
     }
 }
 
+
 /* Resolve OpenMP directive clauses and check various requirements
    of each directive.  */
 
